//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#include "Ensure.h"
// cwz-build-command: cd ../MasterControl; make all

double AdvSkills::goToThresh(double error, double tdist)
{
  // check if turn is needed (distance depended)
  // visual angle of target/error circle
  double gamma = atan2(error, tdist);
  double split = 1/(1 + goToThreshStretch*(tdist - error));
  return (1 - split)*M_PI/2 + split*gamma;
}

AdvSkills::TurtleCommand
AdvSkills::goToPosition(const Snapshot *sn,
			const Vector2d &target,
			double error, double stamina,
			bool minAngle, Angle prefAngle)
{
  TurtleCommand com;
  com.state = TONWAY;
  // time information not valid
  com.time = -1;
  // simple turn/dash algorithm. Will not work for small errors, I guess.
  // But it is not for that at all;

  com.v = target - sn->my.pos; 
  Angle tangle = com.v.getAngle();
  double tdist = norm(com.v);

  if (tdist <= error) {
    com.time  = 0;
    com.state = TFINISHED;
    return com;
  }

  // angle minimization if wanted
  bool reverse = false;
  if (minAngle && (fabs(tangle - prefAngle) > M_PI/2)) {
    reverse = true;
    tangle += PI;
    com.state = TGOINGBACKWARDS;
  }

  // check if turn is needed (distance depended)
  Angle diff = tangle - sn->my.bodyDir;
  if (fabs(diff) > goToThresh(error, tdist)) {
    com.v       = sn->my.v;
    TurnCommand tmptcm = turn(sn, tangle);
    com.bodyDir = tmptcm.bodyDir;
    com.turnMoment = tmptcm.turnMoment;
    com.command = TurtleCommand::TURN;
    return com;
  }

  double minacc, maxacc;
  calcAcc(sn, stamina, minacc, maxacc);

  double l1, l2, tgtDash = 0;
  // overkill since this is very likely more exact than error implies
  switch (intersectLineCircle(radAng2Vector2d(1, sn->my.bodyDir),
			      com.v - sn->my.v, error, l1, l2)) {
  case None: // no panic. We didn't turn since we are far away.
    // dash at full speed (haha)
    if (reverse)
      tgtDash = minacc;
    else
      tgtDash = maxacc;
    
    break;
  case Valid:
    if (l1 > maxacc)
      // dash forward
      tgtDash = maxacc;
    else if (l2 < minacc)
      // dash backwards
      tgtDash = minacc;
    else
      tgtDash = (max(l1, minacc) + min(l2, maxacc))/2;
    break;
  case Instable: // impossible, norm(rad...) = 1
  default: ; // R1larger/R2larger only for Circle/Circle
  }

  com.v       = sn->my.v + radAng2Vector2d(tgtDash, sn->my.bodyDir);
  com.bodyDir = sn->my.bodyDir;
  com.command = TurtleCommand::DASH;
  return com;
}

double AdvSkills::reachRadius(double dlen, int numDashes, double decay)
{
  return dlen/(1 - decay)*
    (numDashes - decay*(1 - power(decay, numDashes))/(1 - decay));
}

AdvSkills::TurtleCommand
AdvSkills::goToPositionTimed(const Snapshot *sn, const Vector2d &target,
			     double stamina, double error)
{
  WorldMovedObject tgt(0, 0, 0, target, Vector2d(0, 0));
  WorldMovedObject tmp;
  return interceptTimed(sn, tgt, stamina, kickRange, tmp); 
}

AdvSkills::TurtleCommand
AdvSkills::controlBall(const Snapshot *sn, double stamina)
{
  WorldBall tmp;
  return interceptTimed(sn, sn->ball, stamina, kickRange, tmp); 
}

AdvSkills::TurtleCommand
AdvSkills::controlBall(const Snapshot *sn, double stamina,
		       double error, WorldBall &meet)
{
  return interceptTimed(sn, sn->ball, stamina, error, meet);
}

AdvSkills::TurtleCommand
AdvSkills::interceptTimed(const Snapshot *sn, const WorldMovedObject &target,
			  double stamina, double error,
			  WorldMovedObject &meet)
{
#ifdef TRACE_INTERCEPT
  { ostrstream tmp;
  tmp << "stamina input = " << stamina << (char) 0;
  char * str = tmp.str();
  ICDRAWTRACE(MLog(str)); 
  delete str;}
#endif
  TurtleCommand com;
  com.state = TONWAY;

  WorldPlayer simMy(sn->my);
  double bdist = simMy.distTo(target);
  const Vector2d mybvec = radAng2Vector2d(1, sn->my.bodyDir);

  meet = target;

  if (bdist < error) {
    com.time  = 0;
    com.state = TFINISHED;
    return com;
  }

  // we need at least one dash
  meet.simulate();
  simMy.simulate();

  ///////////////////////////////// single dash ?
  // also dashes backwards
  Vector2d relpos = meet.pos - simMy.pos;
  double l1, l2;

  // last dash ALWAYS at full power
  double maxacc = simMy.dashPowerRate * MAXPOWER, minacc = simMy.dashPowerRate * MINPOWER;

  // check for dash on innerSafeCircle, or if not possible, into error
  // so we are as close as still safe to the target
  ICDRAWTRACE(CIRCLE(target.pos, innerSafeCircle));
  if (intersectLineCircle(radAng2Vector2d(1, simMy.bodyDir),
			  relpos, innerSafeCircle,
			  l1, l2) == Valid) {
    double tgtDash = 0;
    bool use = false;
    bool l1ok = (l1 >= minacc) && (l1 <= maxacc);
    bool l2ok = (l2 >= minacc) && (l2 <= maxacc);

    if (l1ok && !l2ok) {
      tgtDash = l1;
      use = true;
    }
    else if (l2ok && !l1ok) {
      tgtDash = l2;
      use = true;
    }
    else if (l1ok && l2ok) {
      // use lower dash (dashing backwards costs double stamina)
      double l1stamina = (l1 > 0) ? l1 : 2*l1;
      double l2stamina = (l2 > 0) ? l2 : 2*l2;
      tgtDash = (fabs(l1stamina) < fabs(l2stamina)) ? l1 : l2;
      use = true;
    }
    if (use) {
      if (l1ok) {
	ICDRAWTRACE(TEXT("min ok", TextPos1));
      }
      if (l2ok) {
	ICDRAWTRACE(TEXT("max ok", TextPos2));
      }

      Vector2d dashVec = radAng2Vector2d(tgtDash, simMy.bodyDir);
      if (tgtDash < 0)
	com.state = TGOINGBACKWARDS;
      com.time    = 1;
      com.v       = sn->my.v + dashVec;
      com.bodyDir = sn->my.bodyDir;
      com.command = TurtleCommand::DASH;

      ICNAMEDLINE("my v", sn->my.pos, sn->my.v);
      ICNAMEDLINE("dash only safe circle", simMy.pos, dashVec);
      ICDRAWTRACE(CIRCLE(meet.pos, innerSafeCircle));

      return com;
    }
  }

  // safeCircle didn't do it, try given error
  // because of the above intersection, we know that
  // our result can't end up with a collision
  ICDRAWTRACE(CIRCLE(target.pos, error));
  if (intersectLineCircle(radAng2Vector2d(1, simMy.bodyDir),
			  relpos, error,
			  l1, l2) == Valid) {
    if ((l1 <= maxacc) && (l2 >= minacc)) {
      double wacc = (l1 + l2)/2;
      if (wacc > maxacc)
	wacc = maxacc;
      if (wacc < minacc)
	wacc = minacc;
      Vector2d dashVec = radAng2Vector2d(wacc, simMy.bodyDir);      
      if (wacc < 0)
	com.state = TGOINGBACKWARDS;
      com.time    = 1;
      com.v       = sn->my.v + dashVec;
      com.bodyDir = sn->my.bodyDir;
      com.command = TurtleCommand::DASH;

      ICNAMEDLINE("my v", sn->my.pos, sn->my.v);
      ICNAMEDLINE("dash only kick range", simMy.pos, dashVec);
      ICDRAWTRACE(CIRCLE(meet.pos, error));

      return com;
    }
  }

  // now we need at least two steps
  meet.simulate();
  simMy.simulate();

  ///////////////////////////////// turn-dash ?
  relpos = meet.pos - simMy.pos;
  Vector2d vl1, vl2, dashvec(0, 0);
  bool tdgood = false;
  switch(intersectCircleCircle(maxacc, relpos, error, vl1, vl2)) {
  case Valid:
    tdgood = true;
    dashvec = 0.5*(vl1 + vl2);
    break;
  case R1larger:
    tdgood = true;
    dashvec = relpos;
    break;
  case R2larger: // hmm, we already got the ball (???)
    tdgood = true; // do essentially nothing (since we turn fist)
    dashvec = radAng2Vector2d(0.0001, relpos.getAngle());
  case None:
  default:;
  }

  if (tdgood) {
    ICNAMEDLINE("turn dash", simMy.pos, dashvec);
    com.time    = 2;
    com.v       = sn->my.v;
    com.command = TurtleCommand::TURN;

    TurnCommand tmptcm = turn(sn, dashvec.getAngle()); 
    com.bodyDir = tmptcm.bodyDir;
    com.turnMoment = tmptcm.turnMoment;
    if (tmptcm.state != RANGLE)
      return com;
    // check if proposed turn still is enough
    double l1, l2;
    if (intersectLineCircle(radAng2Vector2d(1, tmptcm.bodyDir),
			    relpos, error, l1, l2) == Valid) {
      ICDRAWTRACE(TEXT("angle failure, but still ok forwards", TextPos3));
      ICDRAWTRACE(LINE(simMy.pos + l1*dashvec, simMy.pos + l2*dashvec));
      // there IS a solution with that angle
      if ((l1 <= maxacc) && (l2 >= minacc))
	return com;
      
    }

    // it's not as easy. Maybe turning into the opposite dir is ok (for dashing backwards)
    // same process again
    tmptcm = turn(sn, PI + dashvec.getAngle()); 
    com.bodyDir = tmptcm.bodyDir;
    com.turnMoment = tmptcm.turnMoment;
    if (tmptcm.state != RANGLE)
      // HA!!! THAT works (sorry for the stamina)
      return com;
    // check if proposed turn still is enough
    if (intersectLineCircle(radAng2Vector2d(1, tmptcm.bodyDir),
			    relpos, error, l1, l2) == Valid) {
      ICDRAWTRACE(TEXT("angle failure, but still ok backwards", TextPos3));
      ICDRAWTRACE(LINE(simMy.pos + l1*dashvec, simMy.pos + l2*dashvec));
      // there IS a solution with that angle
      if ((l1 <= maxacc) && (l2 >= minacc))
	return com;
    }
  }

  ///////////////////////////////// dash-dash ?
  double dpower = simMy.dashPowerRate * min(MAXPOWER, (simMy.staminaIncMax + stamina/2));
  double len = reachRadius(dpower, 2, simMy.decay);

  
  ICDRAWTRACE(POSITION(simMy.pos + len*mybvec));

  double err = error + bdist*(exp(QUANTIZE_STEP) - 1);
  if ((intersectLineCircle(mybvec, meet.pos - simMy.pos, err, l1, l2) == Valid) &&
      (l1 < len) && (l1 > 0)) {
    // I can get there without turn
    // l1 is the needed dash fro a one cycle hit (impossible)
    // cut down if more dashes are needed
    if (l1 > dpower)
      l1 = dpower;
    com.time    = 2;
    com.v       = sn->my.v + radAng2Vector2d(l1, simMy.bodyDir);
    com.bodyDir = sn->my.bodyDir;
    com.command = TurtleCommand::DASH;

    ICDRAWTRACE(CIRCLE(meet.pos, err));
#ifdef TRACE_INTERCEPT
    char buf[256];
    sprintf(buf, "dashing straight, length=%f, time=2", len);
    ICDRAWTRACE(TEXT(buf, TextPos));
#endif
    return com;
  }

  ///////////////////////////////// simulation, t >= 3
  com.state = TONWAY;

  // after this loop, tgt is where we head
  // initialize to infinity position
  Vector2d tgt = target.pos + (1/(1 - target.decay))*target.v;
  dpower       = simMy.dashPowerRate * simMy.staminaIncMax;

  int time = 2; // this we know from before
  // hopefully not an endless loop
  // preserve stamina for last turn with full power
  stamina = max(0.0, stamina - 2*MAXPOWER);

  for (;;) {
    time++;
    meet.simulate();
    simMy.simulate();

    // first check if dashing alone would do it
    dpower = simMy.dashPowerRate * min(MAXPOWER, (simMy.staminaIncMax + stamina/time));
    len = reachRadius(dpower, time, simMy.decay);

    ICDRAWTRACE(POSITION(simMy.pos + len*mybvec));

    double err = error + bdist*(exp(QUANTIZE_STEP) - 1);
    double l1, l2;
    ICDRAWTRACE(CIRCLE(meet.pos, err));
    if ((intersectLineCircle(mybvec, meet.pos - simMy.pos, err, l1, l2) == Valid) &&
	(l1 < len) && (l1 > 0)) {
      // I can get there without turn
      // l1 is the needed dash fro a one cycle hit (impossible)
      // cut down if more dashes are needed
      if (l1 > dpower)
	l1 = dpower;
      com.time    = time;
      com.v       = sn->my.v + radAng2Vector2d(l1, sn->my.bodyDir);
      com.bodyDir = sn->my.bodyDir;
      com.command = TurtleCommand::DASH;

#ifdef TRACE_INTERCEPT
      char buf[256];
      sprintf(buf, "dashing straight, length=%f, time=%d", l1, time);
      ICDRAWTRACE(TEXT(buf, TextPos));
#endif
      return com;
    }

    // now check one turn, if possible
    dpower = simMy.dashPowerRate * min(MAXPOWER, (simMy.staminaIncMax + stamina/(time - 1)));
    len = reachRadius(dpower, time - 1, simMy.decay);

    ICDRAWTRACE(CIRCLE(simMy.pos, len));

    Vector2d vl1, vl2;
    bool usetgt = false;
    switch (intersectCircleCircle(len, meet.pos - simMy.pos, innerSafeCircle, vl1, vl2)) {
    case Valid: // goto into middle of solutions
      tgt = simMy.pos + 0.5*(vl1 + vl2);
      usetgt = true;
      break;
    case R1larger:
      // go straight into target
      tgt = meet.pos;
      usetgt = true;
    default:;
    }

    if (usetgt) {
      // check for angle not too large
      Angle dashdir = (tgt - simMy.pos).getAngle();
      TurnCommand tcm = turn(sn, dashdir);
      // works, go to tgt
      if (tcm.state != RANGLE)
	break;
    }

    // check two turns
    dpower = simMy.dashPowerRate * min(MAXPOWER,
				       (simMy.staminaIncMax + stamina/(time - 2)));
    len = reachRadius(dpower, time - 2, simMy.decay);

    ICDRAWTRACE(CIRCLE(simMy.pos, len));

    usetgt = false;
    switch (intersectCircleCircle(len, meet.pos - simMy.pos, innerSafeCircle, vl1, vl2)) {
    case Valid: // goto into middle of solutions
      tgt = simMy.pos + 0.5*(vl1 + vl2);
      usetgt = true;
      break;
    case R1larger:
      // go straight into target
      tgt = meet.pos;
      usetgt = true;
    default:;
    }

    if (usetgt)
      // ok, with two turns we get the ball, go!
      break;
  }

  // if we fall here, tgt is where we should TURN
  // dashing straight is handled above
  // determine wether we should turn. This is a crude approximation
  ICDRAWTRACE(CIRCLE(meet.pos, BALL_SIZE));

  Angle dashdir = (tgt - simMy.pos).getAngle();
  TurnCommand tcm = turn(sn, dashdir);
  com.turnMoment = tcm.turnMoment;
  com.bodyDir = tcm.bodyDir;
  com.v       = sn->my.v;
  com.time    = time;
  com.command = TurtleCommand::TURN;

  ICDRAWTRACE(LINE(simMy.pos, tgt));
#ifdef TRACE_INTERCEPT
  char buf[256];
  sprintf(buf, "turning, time=%d", time);
  ICDRAWTRACE(TEXT(buf, TextPos));
#endif
  return com;
}

void AdvSkills::calcAcc(const Snapshot *sn, double stamina,
			double &minacc, double &maxacc)
{
  // ok, dash
  if (stamina < 0)
    stamina = 0;

  maxacc = min(stamina, MAXPOWER) * sn->my.dashPowerRate;
  minacc = max(-0.5*stamina, MINPOWER) * sn->my.dashPowerRate;

  double l1, l2;
  // max accelerations not exceeding player's maximum velocity
  switch (intersectLineCircle(radAng2Vector2d(1, sn->my.bodyDir),
			      -sn->my.v, sn->my.vmax, l1, l2)) {
  case Valid:
    if ((l1 > minacc) && (l1 <= maxacc))
      minacc = l1;
    if ((l2 < maxacc) && (l2 >= minacc))
      maxacc = l2;
    break;
  case None: // Currently too fast (???)
  case Instable: // 1 == 0 (????)
  default:;
  }
}

void AdvSkills::exec(const Snapshot *sn, const AdvSkills::TurtleCommand &com) {
  if (com.state == TFINISHED) {
    cerr << sn->time << "AdvSkills::exec(TurtleCommand): called when finished" << endl;
    noAction();
    return;
  }
  switch (com.command) {
  case TurtleCommand::TURN:
    basicTurn(com.turnMoment);
    
    break;
  case TurtleCommand::DASH: {
    // determine sign of direction
    double pow = norm(com.v - sn->my.v)/sn->my.dashPowerRate;
    if (com.state == TGOINGBACKWARDS)
      pow = -pow;
    basicDash(pow);
    break;
  }
  default:
    cerr << sn->time << "AdvSkills::exec(TurtleCommand): called without valid command" << endl;
    cerr << sn->time << "AdvSkills::exec(TurtleCommand): time = " << com.time << endl; 
    cerr << sn->time << "AdvSkills::exec(TurtleCommand): com  = " << (int) com.command << endl; 
    cerr << sn->time << "AdvSkills::exec(TurtleCommand): comv = " << com.v << endl; 
    ensure(0);
    noAction();
  }
}
