//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
//
// libMRB Comm.h 
//
// $Id: Comm.h,v 1.10 2002/05/09 14:10:20 cs Exp $
// $manager: cs$
//
//////////////////////////////////////////////////////////////////////

#ifndef COMM_H
#define COMM_H

#include "UDPSocket.h"
#include "mrbtypes.h"
#include "constants.h"
#include "sensor.h"
#include "datastore.h"
#include "asyncinfo.h"
#include "audiomsg.h"
#include "Parser.h"
#include "audiomgr.h"
#include "mcommand.h"

/// error types returned by Comm
enum comm_error {
  COMM_OK = 0,
  COMM_ERROR,	           	///< unspecified error
  COMM_ERROR_SERVER_NOT_FOUND,  ///< server name not found
  COMM_ERROR_SOCKET_INIT,	///< couldn't init socket
  COMM_ERROR_SERVER_INIT,	///< server didn't accept our init command
  COMM_ERROR_NO_DATA,	        ///< server has no new data
  COMM_ERROR_PARSE,	        ///< server sent us something we could not parse
};

/// This is the class which handles all communication of the agent to and from the soccerserver.
/** The main function of the Comm is the communication with the soccerserver. It keeps track of
    the number of commands sent and resends them if they are lost. Therefore you have to call 
    handleLostCmds(). The main function of the Comm is drainSocket() which reads data from the
    soccerserver as long as data is available or the agent should calculate a command for the
    current cycle.
 */
class Comm {
public:
  /** Constructor of the Comm. You can only have one Comm per agent.
      @param team Name of the team
      @param ctype The type of the client (coach or player)
      @param host The name of the server
      @param port The port of the server
      @param reconnectNo The number of the player reconnecting (-1 = no reconnect)
      @param FUNSSELMode true to be able to send MCommands
      @param clangMinVer the minimal coachlanguage version supported
      @param clangMaxVer the maximal coachlanguage version supported
      @param coachSide The side of the coach (only important for an offlinecoach)
      @see mcommand.h
  */
  Comm (const char * team, client_type ctype,
	const char * host, int port, int reconnectNo = -1,
	bool FUNSSELmode = false, int clangMinVer = 7, int clangMaxVer = 7,
	playing_side coachSide = SIDE_LEFT );
  /// The destructor
  ~Comm( void );

  /** @name Interface to Player and Coach  */
  //@{
  /// Connect to server.
  /** Connect to server specified in Constructor. 
      That is check in at server, and parse response.
      @see comm_error
   */
  comm_error connect( void );
  
  inline bool isServerAlive( void ) { return serverOK; };
  inline bool isGameOver( void ) { return (async->getPlayMode() == PLAY_MODE_TIME_OVER); };
  inline bool isTimeRunning( void ) { return async->isTimeRunning(); };
  inline bool isActionNeeded( void ) { return async->actNow(); };
  inline bool inFUNSSELmode( void ) { return FUNSSELmode; };
  inline int getServerTime( void ) { return async->getServerTime(); };

  inline int getMyNumber( void ) { return async->getNumber(); };
  inline bool amIGoalie( void ) { return async->isGoalie(); };
  inline const char * getMyTeam( void ) { return async->getTeam(); };
  inline playing_side getMySide( void ) { return async->getSide(); };

  inline play_mode getPlayMode( void ) { return async->getPlayMode(); };
  inline int getOurScore( void ) { return async->getOurScore(); };
  inline int getOtherScore( void ) { return async->getOtherScore(); };
  inline bool inSecondHalf( void ) { return async->inSecondHalf(); };
  
  void say( AudioMsg & msg );

  // This is for sending a string directly (soccerserver 8.* hack)
  void say_10( char * msg );

  /// This should be used when sending messages in coachlanguage
  /** This method is not implemented yet
   */
  void coachSay( void /*???*/ );
  inline bool getNextMsg( AudioMsg & msg ) { return audiomgr->getNextMsg(msg); };

  /// Use this for sending draw, log, etc. commands to FUNSSEL
  /** This method is static as no module has a pointer to the Comm.
      It only can be used when there is a connected comm!
   */
  static void  sendMCmd( const MCommand & cmd );
  //@}

  /** @name Action interface */
  //@{
  void noAction( void );
  void sendKick( double power, double dir_rad );
  void sendDash( double power );
  void sendTurn( double moment_rad );
  void sendCatch( double dir_rad );
  void sendMove( double x, double y );

  void sendTurnNeck( double angle_rad );
  /** send change view command to sserver
      If send is false the method returns the number of cycles to wait before the command
      should be sent, in order to get the see messages as early as possible in a cycle
      @param width The view angle the agent wants to get (wide, normal, narrow)
      @param quality The quality the agents wants to get (hight, low)
      @param send if true the change view command is sent otherwise it is not sent
      @return Number of cycles to wait before the command should be sent (0 if send == true)
   */
  void sendChangeView( view_width width, view_quality quality );
  void sendPointTo( double dist, double dir_rad );
  void sendPointToOff( void );
  void sendAttentionTo( int pNum, bool ourTeam = true );
  void sendAttentionToOff( void );
  void sendTackle( double power );
  void sendSenseBody( void );
  void sendSay( const char * message );
  void sendDone( void );

  void sendChangePlayer( int player, int type );
  void sendChangeMode( play_mode playMode );
  void sendMoveObject( const char * object, double x, double y );
  void sendMoveObject( const char * object, double x, double y, double dir_rad );
  void sendMoveObject( const char * object, double x, double y, double dir_rad, 
		       double xvel, double yvel );
  void sendEye( bool eyeOn );
  void sendEar( bool earOn );
  void sendLook( void );  
  void sendCheckBall( void );
  void sendRecover( void );
  //@}

  /** @name Interface to other modules (worldmodel, replayer) */
  //@{
  inline bool getNextData( SensorData &data ) { return store->getNext(data); };
  inline void getLastAction( CycleAction &action ) { async->getLastAction(action); };

  void drainSocket( void );
  //@}

private:

  UDPSocket * sock;
  Parser * parser;
  SensorDataStore * store;
  AsyncCommInfo * async;
  AudioMgr * audiomgr;

  bool FUNSSELmode;
  bool serverOK;
  int reconnectNumber;
  client_type ctype;
  playing_side coachSide;

  static Comm * theComm;

  comm_error checkinAtServer( void );
  void generateCheckinMessage( char * outmsg );
  comm_error parseCheckinResponse( const char * inmsg );
};
#endif
