//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#ifndef UDPSOCKET_H
#define UDPSOCKET_H

#include <string.h>
#include <sys/types.h>
#include <arpa/inet.h>

/// Error thrown by UDPSocket
class UDPSocketError {
public:
  /// create a new error
  /** @param msg Description of the error
      @param errNo Number of the error
  */
  UDPSocketError( const char * msg, int errNo );

  /// copy constructor
  UDPSocketError( const UDPSocketError & );

  /// clean up
  ~UDPSocketError( void );

  /// the message to describe the error
  const char * msg( void );

  /// the number of the error
  int number( void );

private:
  char * errmsg;
  int errNo;
};

/// special error thrown by recv
class UDPTimeOut : public UDPSocketError {
public:
  UDPTimeOut( void );
  UDPTimeOut( const UDPTimeOut & );
  ~UDPTimeOut( void );
};

/// special error thrown by recv
class UDPAct : public UDPSocketError {
public:
  UDPAct( void );
  UDPAct( const UDPAct & );
  ~UDPAct( void );
};

/// Basic UPD Socket used by the Comm
/** This is a blocking UDP socket, which uses select to have a timeout
    for reading. When select returns with a value of -1 recv throws an
    UDPSocketError with number set to errno. i.e. When an interrupt occurs
    number is set to EINTR.
 */
class UDPSocket {
public:
  /** Constructor used to set up the socket and bind it to a port.
      @param host The name of the serverhost to connect to
      @param port The port of the server at host
      @param maxMsgSize The max. size of the message being awaited
   */
  UDPSocket( const char * host, int port, size_t maxMsgSize )
    throw (UDPSocketError);

  /// clean up
  /** close the socket and free memory */
  ~UDPSocket( void );
  
  /// send a message to the server
  /** This method will block till the message can be sent or an error occurs.
      @param msg The message to be sent to the server
   */
  void send( const char * msg )
    throw (UDPSocketError);

  /// wait till there is a message from the server
  /** Wait till there is a message and store it local.
      @param msec Wait at most that manny milliseconds. If sec < 0
      recv will wait forever if nothing arrives.
      @param act If != NULL and true UDPAct will be thrown before blocking in select
      @return Pointer to the message recved from server.
      This msg will be overwritten with the next msg recvd.\\
      If nothing arrived msg[0] == 0.
   */
  const char * recv( int msec, const bool * act = NULL )
    throw (UDPSocketError, UDPTimeOut, UDPAct);

private:
  /// socket descriptor
  int sd;
  /// adress structure of server
  sockaddr_in srv;
  /// msg recvd from server
  char * recMsg;
  /// max len of msg to be recvd
  size_t maxMsgSize;
};
#endif




