//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

// -*- c++ -*-/////////////////////////////////////////////////////////
//
// $Id: coach.h,v 1.15 2001/09/20 15:16:09 satan Exp $
//
// coach.h - interface for training coach client
//
///////////////////////////////////////////////////////////////////////


#ifndef _COACH_H_
#define _COACH_H_

#include <unistd.h>
#include <iostream.h>
#include <fstream.h>
#include <strstream.h>

#include "constants.h"
#include "Comm.h"
#include "worldmodel.h"
#include "audiomgr.h"

// filename of snapshot log files
#define SNAP_LOG_FILENAME "snaps_"
#define SNAP_LOG_EXTENSION ".log"


class Coach {

public:


  // Constructor / Destructor
  ///////////////////////////////////////////////////////////////////////////

  // Generates and logs in a training coach client at the given soccerserver.
  // After the constructor has finished all properties of the agent are set.
  // 
  // Call this constructor first in your derived classes constructors before
  // you implement any further initialization of the agent.
  //
  // Parameter description:
  // - numSnapshots
  // Specifies the memory usage of the Worldmodel. Larger numbers will keep
  // more snapshots of the past accessible, but consume quite some memory.
  Coach (const char*  team_name,             // name of own team
	 playing_side side,                  // side on which we are playing
	 const char*  host         = DEFAULT_HOST,
	 int          port         = COACH_PORT,   
	 const char*  configfile   = "agent.conf",
	 bool         logging_snap = false,  // logs all snapshots
	 bool         FUNSSELmode  = false,  // tells if we can send mcommands
	 int          numSnapshots = DEFAULT_SNAPSHOT_DEPTH); // see below


  // Starts execution of the Coach client. 
  // After calling this method, the agent will repeatedly call act(snap)
  // once every timestep to perform the coaching.
  //
  // You cannot override this method.
  void coach();
  
  
  // Called once every timestep if game timer is running, otherwise once every
  // SIMULATOR_STEP ms.
  //
  // Override this method virtual in your derived classes to implement 
  // coach behaviour.
  virtual void act(Snapshot* snap) = 0;
  
  
  // Cleans up and logs out
  //
  // You can put own code in your derived destructors. As the final action call
  // this destructor so the agent will leave the simulation properly 
  // and release its resources (files, sockets, windows, etc).
  virtual ~Coach();  
  

  
  // general information about the client himself and the server state
  /////////////////////////////////////////////////////////////////////

  // obvious
  int          getMyNumber()  {return comm->getMyNumber();} // is 0 for coach
  const char*  getMyTeam()    {return comm->getMyTeam();}
  playing_side getMySide()    {return comm->getMySide();}
  play_mode    getPlayMode()  {return comm->getPlayMode();}
  
  // training coach actions
  void         changePlayerType(int player, int type);
  void         changeMode(play_mode pm);
  void         ear(bool earOn);         // enables/disables hearing
  void         moveObject (const char *object, double x, double y);
  void         moveObject (const char *object, double x, double y, 
			   double dir_rad);
  void         moveObject (const char *object, double x, double y, 
			   double dir_rad, 
			   double xvel, double yvel);
  void         recover(); // resets all players stamina and effort
  
  // returns the timestep the server is in, our current time.
  // (There is only ONE current time. This is it.)
  int          getTime()      {return comm->getServerTime();} 
  
  // if needed, returns a snapshot for a given time from the worldmodel
  Snapshot*    getSnapshot(int time) {return world->getSnapshot(time);}

  // sends an audiomessage to other clients immediately
  void         say(AudioMsg msg)         {comm->say(msg);}

  // returns true, if a new audiomessage is available in msg, false otherwise
  bool         getNextMsg(AudioMsg &msg) {return comm->getNextMsg(msg);}

  
private:

  Comm *comm;

  Worldmodel *world;

  bool writeSnaps;
  ofstream out;

};


#endif
