//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

// -*- c++ -*- ///////////////////////////////////////////////////////
//
// mcommand.h - some classes to send commands to special monitor
//              controlling the game
//
// $Id: mcommand.h,v 1.11 2001/09/11 10:01:38 goetz Exp $
// Time-stamp: <2001-04-24 18:42:55 goetz>
//////////////////////////////////////////////////////////////////////

#ifndef MCOMMAND_H
#define MCOMMAND_H

#include "coordinates.h"
#include <string>

/** base class
 * duration in subclasses means number of steps (how long the command
 * should be executed (default 1 means only current timestep!)
 */
class MCommand
{
protected:
  string encoded;
public:
  const int duration;

  operator const string &() const { return encoded; }
  int getLength()           const { return encoded.length(); }

  /** default contructor for creating empty component (with duration)
   */
  MCommand( int duration = 0) : duration(duration) {};
  MCommand( int duration, const string &encoded); 
};

/** output operator for debugging
 */
class ostream& operator<< (class ostream&, const MCommand &);

/** marker (kind of draw command, but can be searched for!)
 */
class MMarker : public MCommand
{
public:
  MMarker( const Vector2d &, const char *label= 0,int duration = 0);
};

/** logfile entry
 */
class MLog : public MCommand
{
public:
  MLog( const char *, int duration = 0);
};

/** worldmodel snapshot
 */
class MSnapshot : public MCommand
{
public:
  // duration 0, nothing else makes sense!
  MSnapshot( const class Snapshot & );
};

// style for lines (used in Circles, Lines, Boxes)
enum MStyle { MSolid=0, MDash=1, MDot=2, MDashDot=3, MDashDotDot=4};

/** a circle
 * with center "pos", radius "rad" and maybe gridlines marking the center
 */
class MDCircle : public MCommand
{
public:
  MDCircle(const Vector2d &pos, double rad = 1, bool hasGrid = false,  
	   MStyle style= MSolid, int duration = 0);
};

/** a position marker
 */
class MDPosition : public MCommand
{
public:
  MDPosition(const Vector2d &pos, int duration = 0);
};

/** a line from startpoint to endpoint
 */
class MDLine : public MCommand
{
public:
  MDLine(const Vector2d &start, const Vector2d &end, MStyle style= MSolid,
	 int duration = 0);
};

/** a vertical line through x coordinate "x"
 */
class MDVLine : public MCommand
{
public:
  MDVLine(double x, MStyle style= MSolid, int duration = 0);
};

/** a box from startpoint to endpoint
 */
class MDBox : public MCommand
{
public:
  MDBox(const Vector2d &start, const Vector2d &end, MStyle style= MSolid,
	int duration = 0);
};

/** a corridor as described in "snapshot.h"
 */
class MDCorridor : public MCommand
{
public:
  MDCorridor( const Vector2d &source, const Vector2d &target,
	      double diam, const Angle &alpha,
	      MStyle style= MSolid, int duration = 0);
};

/** Labels to label objects on the field
 */
class MDLabel : public MCommand
{
public:
  MDLabel( const Vector2d &pos, const string &text, int duration = 0 );
};

/** simple stop command
 */
class MStop : public MCommand
{
public:
  MStop();
};

/** commands like the foul menu in the soccerserver
 */

/** drop ball command
 */
class MDropBall : public MCommand
{
public:
  MDropBall( const Vector2d &pos );
};

/** free kick us command
 */
class MFreekickUs : public MCommand
{
public:
  MFreekickUs( const Vector2d &pos );
};

/** drop ball command
 */
class MFreekickThem : public MCommand
{
public:
  MFreekickThem( const Vector2d &pos );
};

/** set a player to a certain place (monitor command!)
 */
class MSetPlayer : public MCommand
{
public:
  MSetPlayer( bool ourTeam, int playerNo, const Vector2d &pos, const Angle &ang=0);
};

#endif
